"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) if (e.indexOf(p[i]) < 0)
            t[p[i]] = s[p[i]];
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable indent */
/**
 * @file Compiler
 * @author Cuttle Cong
 * @date 2018/3/25
 * @description
 */
const _ = require("lodash");
const nps = require("path");
const pReduce = require("p-reduce");
const core_1 = require("../../types/core");
const toArray_1 = require("../../lib/toArray");
const match_1 = require("../../lib/match");
const Variables_1 = require("./Variables");
const hookify_1 = require("./hookify");
const matchMeta_1 = require("./matchMeta");
const parseQueryString_1 = require("../../lib/parseQueryString");
const DefaultLogger_1 = require("../DefaultLogger");
const EdamError_1 = require("../EdamError");
const debug = require('debug')('edam:Compiler');
class Compiler extends core_1.AwaitEventEmitter {
    constructor({ loaders, mappers, hookCwd, root } = {}) {
        super();
        this.logger = new DefaultLogger_1.default();
        this.root = '';
        this.hookCwd = process.cwd();
        this.assets = {};
        this.loaders = {
            module: require('./loaders/module'),
            LoDash: [[require('./loaders/lodash'), {}]],
            hbs: [[require('./loaders/plopHandlebar'), {}]]
        };
        this.mappers = [
            {
                test: '*',
                loader: ['LoDash', 'hbs']
            }
        ];
        this.variables = new Variables_1.default();
        if (hookCwd) {
            this.hookCwd = hookCwd;
        }
        if (loaders) {
            this.loaders = loaders;
        }
        if (mappers) {
            this.mappers = mappers;
        }
        if (root) {
            this.root = root;
        }
    }
    removeHook(hookName, hook) {
        this.removeListener(hookName, hook);
        return this;
    }
    addHook(hookName, hook, options = {}) {
        let cmd = typeof hook === 'string' && hook;
        options = Object.assign({ type: 'on', silent: false }, options);
        hook = hookify_1.default(hook, this.hookCwd);
        const wrapped = options.silent
            ? hook
            : (function wrapHook(hook, logger) {
                return function () {
                    logger.log('Trigger hook:', hookName, cmd ? JSON.stringify(cmd) : '');
                    return hook.apply(this, arguments);
                };
            })(hook, this.logger);
        this[options.type](hookName, wrapped);
        return wrapped;
    }
    addLoader(loaderId, loader) {
        this.loaders[loaderId] = toArray_1.default(loader);
        return this;
    }
    removeLoader(loaderId) {
        delete this.loaders[loaderId];
        return this;
    }
    _matchedLoaders(path) {
        let matchedLoader;
        _.some(this.mappers, function (mapper) {
            if (match_1.isMatch(path, mapper.test)) {
                matchedLoader = mapper.loader;
                return true;
            }
        });
        return matchedLoader;
    }
    transform(input, loaders, path, highOrderOptions) {
        return __awaiter(this, void 0, void 0, function* () {
            loaders = toArray_1.default(loaders);
            let data = { input, loaders };
            yield this.emit('loader:each:before', data);
            const result = yield pReduce(data.loaders, (input, loader) => __awaiter(this, void 0, void 0, function* () {
                if (_.isString(loader)) {
                    let { name, query } = parseQueryString_1.default(loader);
                    let id = name;
                    loader = this.loaders[id];
                    if (!loader) {
                        throw new EdamError_1.default(`loaderId: ${id} is not matched.`);
                    }
                    return yield this.transform(input, loader, path, _.isEmpty(query) ? highOrderOptions : query);
                }
                let options = {};
                // [ loader, options ]
                if (_.isArray(loader)) {
                    options = loader[1] || {};
                    loader = loader[0];
                }
                options = _.isEmpty(highOrderOptions) ? options : highOrderOptions;
                const context = yield this.variables.get();
                _.merge(context, {
                    _: {
                        file: {
                            path,
                            dirname: nps.dirname(path),
                            name: nps.basename(path).replace(/(.+)\.[^.]+?$/, '$1'),
                            ext: nps.extname(path)
                        }
                    }
                });
                const loaderSelf = {
                    compiler: this,
                    path,
                    options
                };
                if (loader.raw === true) {
                    input = Buffer.isBuffer(input) ? input : new Buffer(input);
                }
                else {
                    input = String(input);
                }
                try {
                    return yield loader.apply(loaderSelf, [input, context]);
                }
                catch (e) {
                    if (loader.allowError === true) {
                        this.logger.warn('Error occurs when transforming content of file: `' +
                            path +
                            '`, but the loader allows error happening.\n', e);
                        return input;
                    }
                    throw e;
                }
            }), data.input);
            let temp = { result };
            yield this.emit('loader:each:after', temp);
            return temp.result;
        });
    }
    run() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.emit('run:before');
            yield this.emit('assets', this.assets);
            yield this.emit('variables', this.variables);
            const workers = _.map(this.assets, (asset, path) => __awaiter(this, void 0, void 0, function* () {
                const input = asset.value;
                const loaders = asset.loaders;
                const data = { input, loaders };
                let highOrderOptions;
                if (_.isString(data.input)) {
                    const { content, meta } = matchMeta_1.default(data.input, path);
                    data.input = content;
                    if (meta && meta.loader) {
                        debug('matched meta: \n%O \npath: %s', meta, path);
                        if (!(meta.loader.name in this.loaders)) {
                            this.logger.warn('Matched loader %s from file: %s is not existed', meta.loader.name, path);
                        }
                        else {
                            // data.loaders = this.loaders[meta.loader.name]
                            // use name instead of function
                            data.loaders = meta.loader.name;
                            highOrderOptions = meta.loader.query;
                        }
                    }
                }
                if (!data.loaders) {
                    data.loaders = this._matchedLoaders(path);
                }
                debug('loader path: %s, loaders: %o', path, data.loaders);
                if (data.loaders) {
                    try {
                        yield this.emit('loader:before', data);
                        return Object.assign({ path, output: yield this.transform(data.input, data.loaders, path, highOrderOptions) }, data);
                    }
                    catch (err) {
                        this.logger.error('Error occurs when transforming content of file: `' + path + '`\n', err);
                        return Object.assign({ path, error: err }, data);
                    }
                }
                return Object.assign({ path }, data);
            }));
            const array = yield Promise.all(workers);
            let output = {};
            array.filter(Boolean).forEach(data => {
                const { path } = data, rest = __rest(data, ["path"]);
                output[path] = rest;
            });
            yield this.emit('run:after', output);
            return output;
        });
    }
}
exports.default = Compiler;
