"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable indent */
/**
 * @file index
 * @author Cuttle Cong
 * @date 2018/3/25
 * @description
 */
const core_1 = require("../../types/core");
const toArray_1 = require("../../lib/toArray");
const mkdirp = require("mkdirp");
const rimraf = require("rimraf");
const pify = require("pify");
const _ = require("lodash");
const nps = require("path");
const fileSystem_1 = require("../../lib/fileSystem");
const mm = require("micromatch");
const DefaultLogger_1 = require("../DefaultLogger");
const debug = require('debug')('edam:FileProcessor');
const tildify = require('tildify');
class FileProcessor extends core_1.TreeProcessor {
    constructor(tree, dest, emitter) {
        super(tree);
        this.tree = tree;
        this.dest = dest;
        this.emitter = emitter;
        this.logger = new DefaultLogger_1.default();
        this.emitter = emitter || new core_1.AwaitEventEmitter();
        this.dest = dest;
    }
    writeToFile(filepath = this.dest, option = {
        clean: false,
        overwrite: false
    }) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!filepath) {
                filepath = this.dest;
            }
            yield pify(mkdirp)(filepath);
            if (option.clean) {
                yield pify(rimraf)(filepath);
            }
            yield this.emitter.emit('pre', filepath);
            const workers = [];
            _.each(this.tree, (data, path) => {
                const filename = nps.join(filepath, path);
                if (!option.overwrite && fileSystem_1.default.existsSync(filename)) {
                    debug('%s already existed, ignored!', tildify(filename));
                    this.logger.warn('%s already existed, ignored!', tildify(filename));
                    return;
                }
                workers.push((function write() {
                    return __awaiter(this, void 0, void 0, function* () {
                        if (!data.error) {
                            yield pify(mkdirp)(nps.dirname(filename));
                            yield fileSystem_1.default.writeFile(filename, data.output);
                        }
                    });
                })());
            });
            yield Promise.all(workers);
            yield this.emitter.emit('post', filepath);
            return true;
        });
    }
    get(name) {
        return this.tree[name];
    }
    match(m) {
        // const arr =
        return mm(Object.keys(this.tree), m, { dot: true });
    }
    delete(paths) {
        const fileStates = toArray_1.default(paths).map((path) => {
            const state = this.tree[path];
            delete this.tree[path];
            return state;
        });
        return fileStates.length === 1 ? fileStates[0] : fileStates;
    }
    remove(m = []) {
        debug('remove input: %o', m);
        toArray_1.default(m).forEach(eachm => {
            const paths = this.match(eachm);
            this.delete(paths);
        });
    }
    _move(path, to) {
        debug('move from %s to %s', path, to);
        const state = this.delete(path);
        if (state) {
            this.new(to, state);
            return state;
        }
    }
    _copy(path, to) {
        debug('cp from %s to %s', path, to);
        const state = this.get(path);
        if (state) {
            this.new(to, Object.assign({}, state));
            return state;
        }
    }
    move(m, dest) {
        return this._moveOrCopy(m, dest, this._move.bind(this));
    }
    copy(m, dest) {
        return this._moveOrCopy(m, dest, this._copy.bind(this));
    }
    _moveOrCopy(m, dest, action) {
        debug('move input: %o, %s', m, dest);
        toArray_1.default(m).forEach(eachMatcher => {
            const paths = this.match(eachMatcher);
            // mv *.js [dir][name][ext]
            if (/\[(root|path|name|ext)]/.test(dest)) {
                paths.forEach(from => {
                    let pathObj = nps.posix.parse(from);
                    pathObj.path = pathObj.dir;
                    if (pathObj.path !== '' && !pathObj.path.endsWith('/')) {
                        pathObj.path = pathObj.path + '/';
                    }
                    let realDest = dest.replace(/\[(root|path|name|ext)]/g, (_, $1) => {
                        if (typeof pathObj[$1] === 'string') {
                            return pathObj[$1];
                        }
                        return _;
                    });
                    action(from, realDest);
                });
                return;
            }
            // mv a.js b.js
            if (paths.length === 1 && paths[0] === eachMatcher) {
                action(paths[0], dest);
            }
            else {
                // implement simply, not robustly
                // a: mv a/* b
                // b: mv a/ b
                // https://www.npmjs.com/package/micromatch#parse
                const { nodes } = mm.parse(eachMatcher);
                let basename = '';
                let hasGlobFlag = false;
                if (nodes && nodes.length) {
                    nodes.some(function (node) {
                        switch (node.type) {
                            case 'bos':
                            case 'eos':
                            case 'text':
                            case 'slash':
                                basename += node.val;
                                return false;
                            default:
                                hasGlobFlag = true;
                                return true;
                        }
                    });
                }
                // b: mv a/ b
                if (!hasGlobFlag) {
                    basename = '';
                }
                paths.forEach(path => {
                    let to = dest;
                    if (path.startsWith(basename)) {
                        to = nps.join(dest, path.slice(basename.length));
                    }
                    action(path, to);
                });
            }
        });
        return this;
    }
    new(path, file) {
        this.tree[path] = file;
        return this;
    }
}
exports.default = FileProcessor;
