"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const gitClone_1 = require("../../lib/gitClone");
const path_1 = require("path");
const fileSystem_1 = require("../../lib/fileSystem");
const EdamError_1 = require("../EdamError");
const filenamify = require('filenamify');
const npmInstall = require('../../lib/yarnInstall');
const tildify = require('tildify');
const down = require('download');
const c = require('chalk');
const nurl = require('url');
const urlJoin = require('url-join');
// const ora = require('ora')
const debug = require('debug')('edam:pull:git');
function download(url, dest) {
    return down(url, dest, {
        extract: true,
        strip: 1,
        mode: '666',
        headers: { accept: 'application/zip' }
    });
}
function getUrl(source) {
    let newUrl = source.url;
    newUrl = newUrl.replace(/\.git$/, '');
    const { hostname, protocol } = nurl.parse(newUrl);
    if (protocol === 'https:' || protocol === 'http:') {
        if (hostname === 'github.com') {
            newUrl = urlJoin(newUrl, 'archive', (source.checkout || 'master') + '.zip');
        }
        else if (hostname.startsWith('gitlab')) {
            newUrl = urlJoin(newUrl, 'repository/archive.zip?ref=' + (source.checkout || 'master'));
        }
    }
    return newUrl;
}
module.exports = function gitPull(source, destDir, config) {
    return __awaiter(this, void 0, void 0, function* () {
        const { cacheDir, pull: { git, npmClient } } = config;
        const log = (this && this.logger && this.logger.log) || console.log;
        const errorLog = (this && this.logger && this.logger.warn) || console.error;
        function installFromPkg(target) {
            return __awaiter(this, void 0, void 0, function* () {
                debug('installFromPkg: %s', target);
                const path = path_1.join(target, 'package.json');
                if (!fileSystem_1.default.isFile(path)) {
                    debug('installFromPkg: %s does not have the package.json', target);
                    return;
                }
                const text = yield fileSystem_1.default.readFile(path, {
                    encoding: 'utf8'
                });
                const dependencies = JSON.parse(text).dependencies;
                if (dependencies && Object.keys(dependencies).length > 0) {
                    try {
                        yield npmInstall({
                            stdio: 'ignore',
                            respectNpm5: npmClient === 'npm',
                            cwd: target,
                            production: true
                        });
                    }
                    catch (err) {
                        throw new EdamError_1.default(`Install package from git "${source.url}" failed \n` +
                            err.stack);
                    }
                    log('Installed dependencies automatically, depends %s packages.', c.white(Object.keys(dependencies).length));
                }
            });
        }
        let target = path_1.join(destDir, 'git', git, filenamify(source.url));
        function downloadSource() {
            return __awaiter(this, void 0, void 0, function* () {
                const where = path_1.join(target, filenamify(source.checkout || 'master'));
                if (!cacheDir) {
                    yield fileSystem_1.default.cleanDir(where);
                }
                else if (fileSystem_1.default.existsSync(where)) {
                    // cached
                    log('Using cached git assets from %s by %s\n' +
                        'NOTE: Maybe the cache is out of date. you could set cacheDir to be "false" or pull.git to be "clone".', c.cyan.bold(tildify(where)), c.yellow.bold('Download'));
                    yield installFromPkg(where);
                    return where;
                }
                const url = getUrl(source);
                yield download(url, where);
                yield installFromPkg(where);
                // log('Using cached git assets from %s', c.cyan.bold(tildify(target)))
                return where;
            });
        }
        function gitCloneSource() {
            return __awaiter(this, void 0, void 0, function* () {
                debug('gitCloneSource: target is %s', target);
                if (!cacheDir) {
                    yield fileSystem_1.default.cleanDir(target);
                }
                else if (fileSystem_1.default.existsSync(path_1.join(target, '.git'))) {
                    // cached
                    try {
                        yield gitClone_1.pullForce(target);
                        yield gitClone_1.checkout(source.checkout || 'master', { targetPath: target });
                    }
                    catch (err) {
                        throw new EdamError_1.default(err.message + '\n running in ' + tildify(target));
                    }
                    log('Using cached git assets from %s', c.cyan.bold(tildify(target)));
                    yield installFromPkg(target);
                    return target;
                }
                yield gitClone_1.default(source.url, target, {
                    checkout: source.checkout || 'master'
                });
                yield installFromPkg(target);
                return target;
            });
        }
        // destDir/git/url/(clone|download/checkout)
        switch (git) {
            case 'clone':
                try {
                    return yield gitCloneSource();
                }
                catch (error) {
                    errorLog('Error occurs when git clone %s: \n%s\n' +
                        'Fallback to use git download.', source.url, error.message);
                    return yield downloadSource();
                }
            case 'download':
                try {
                    return yield downloadSource();
                }
                catch (error) {
                    errorLog('Error occurs when git download %s: %s\n' +
                        'Maybe the repo of checkout is not found, please check it.\n' +
                        'Fallback to use git clone.', getUrl(source), error.message);
                    return yield gitCloneSource();
                }
            default:
                throw new EdamError_1.default(`Git pull type "${git}" is not allowed, please use one of 'clone' | 'download'`);
        }
    });
};
